import json
from pathlib import Path
from unittest.mock import MagicMock

from fan_tools.python import cache_async, memoize


class TestCache:
    async def test_01_simplest(self, tmp_path):
        fname = tmp_path / 'cache.json'
        model = MagicMock()

        model.json.return_value = '{"a": "b"}'

        @cache_async[type(dict)](fname, model, {})
        async def func():
            return model

        assert not fname.exists()
        await func()
        assert fname.exists()
        assert fname.read_text() == '{"a": "b"}'

        func.reset_cache()
        assert not fname.exists()

    def test_02_unlink(self, tmp_path):
        fname = tmp_path / 'cache.json'
        model = MagicMock()

        model.json.return_value = '{"a": "b"}'

        @cache_async[type(dict)](fname, model, {})
        async def func():
            return model

        assert not fname.exists()
        func.reset_cache()


class TestMemoize:
    memoized = '{"ab82775f76a7c30b94db389aa2e8702f": "{\\"args\\": [\\"a\\"], \\"kwargs\\": {\\"name\\": \\"b\\"}}"}'

    def test_01_basic(self, tmp_path):
        fname = tmp_path / '.cache/memoized_test.json'

        @memoize(fname)
        def func(*args, **kwargs):
            return json.dumps(
                {
                    'args': args,
                    'kwargs': kwargs,
                }
            )

        assert not fname.exists()
        func('a', name='b')
        assert fname.exists()
        assert fname.read_text() == self.memoized

        func.reset_cache()
        assert not fname.exists()

    def test_02_no_params(self):
        fname = Path('.cache/memoized.json')
        if fname.exists():
            fname.unlink()

        @memoize
        def func(*args, **kwargs):
            return json.dumps(
                {
                    'args': args,
                    'kwargs': kwargs,
                }
            )

        assert not fname.exists()
        func('a', name='b')
        assert fname.exists()
        assert fname.read_text() == self.memoized

        func.reset_cache()
        assert not fname.exists()



import pathlib
import zipfile
from io import BytesIO

import pytest
from fan_tools.file_utils import force_zip, merge_many_zip, merge_zip


@pytest.fixture(scope='session')
def zip1():
    zip_buffer = BytesIO()
    with zipfile.ZipFile(zip_buffer, 'a') as zip_file:
        zip_file.writestr('conflict_file', 'value_from_1_arch')
        zip_file.writestr('file_from_1_archive', 'value_for_file_from_1_archive')
    return zip_buffer


@pytest.fixture(scope='session')
def zip2():
    zip_buffer = BytesIO()
    with zipfile.ZipFile(zip_buffer, 'a') as zip_file:
        zip_file.writestr('conflict_file', 'value_from_2_arch')
        zip_file.writestr('file_from_2_archive', 'value_for_file_from_2_archive')
    return zip_buffer


@pytest.fixture(scope='session')
def zip3():
    zip_buffer = BytesIO()
    with zipfile.ZipFile(zip_buffer, 'a') as zip_file:
        zip_file.writestr('conflict_file', 'value_from_3_arch')
        zip_file.writestr('file_from_3_archive', 'value_for_file_from_3_archive')
    return zip_buffer


@pytest.fixture
def zip2_as_str(tmpdir):
    zip_buffer = tmpdir.join('zip2.zip')
    with zipfile.ZipFile(zip_buffer, 'a') as zip_file:
        zip_file.writestr('conflict_file', 'value_from_2_arch')
        zip_file.writestr('file_from_2_archive', 'value_for_file_from_2_archive')
    return zip_buffer


@pytest.fixture
def zip3_as_pathlike(tmpdir):
    zip_buffer = pathlib.Path(tmpdir.join('zip3.zip'))
    with zipfile.ZipFile(zip_buffer, 'a') as zip_file:
        zip_file.writestr('conflict_file', 'value_from_3_arch')
        zip_file.writestr('file_from_3_archive', 'value_for_file_from_3_archive')
    return zip_buffer


def test_merge_zip(zip1, zip2):
    zip_ = zipfile.ZipFile(merge_zip(zip1, zip2))

    assert zip_.open('conflict_file').read().decode() == 'value_from_2_arch'
    assert zip_.open('file_from_1_archive').read().decode() == 'value_for_file_from_1_archive'
    assert zip_.open('file_from_2_archive').read().decode() == 'value_for_file_from_2_archive'


def test_merge_many_zip(zip1, zip2, zip3):
    zip_ = zipfile.ZipFile(merge_many_zip((zip1, zip2, zip3)))

    assert zip_.open('conflict_file').read().decode() == 'value_from_3_arch'
    assert zip_.open('file_from_1_archive').read().decode() == 'value_for_file_from_1_archive'
    assert zip_.open('file_from_2_archive').read().decode() == 'value_for_file_from_2_archive'
    assert zip_.open('file_from_3_archive').read().decode() == 'value_for_file_from_3_archive'


def test_merge_many_zip_different_types(zip1, zip2_as_str, zip3_as_pathlike):
    zip_ = zipfile.ZipFile(merge_many_zip((zip1, zip2_as_str, zip3_as_pathlike)))

    assert zip_.open('conflict_file').read().decode() == 'value_from_3_arch'
    assert zip_.open('file_from_1_archive').read().decode() == 'value_for_file_from_1_archive'
    assert zip_.open('file_from_2_archive').read().decode() == 'value_for_file_from_2_archive'
    assert zip_.open('file_from_3_archive').read().decode() == 'value_for_file_from_3_archive'


def test_force_zip(zip1):
    res1 = force_zip(zip1)
    assert res1 == zip1

    filename = 'test.json'
    content = b'{"a":"b"}'
    res2 = force_zip(BytesIO(content), filename)

    assert zipfile.is_zipfile(res2)

    zip_obj = zipfile.ZipFile(res2)

    assert len(zip_obj.namelist()) == 1
    assert zip_obj.read(filename) == content



import io
import pathlib
import zipfile

import pytest

from fan_tools.python import root_directory


@pytest.fixture
def simple_file(tmpdir):
    def create(filename: str, content):
        file_dir = tmpdir
        parts = filename.split('/')
        for part in parts[:-1]:
            if pathlib.Path(f'{file_dir}/{part}').exists():
                continue
            file_dir = file_dir.mkdir(part)
        file_dir = file_dir.join(filename.split('/')[-1])
        file_path = pathlib.Path(file_dir)
        mode = 'w'
        if isinstance(content, bytes):
            mode += 'b'
        with open(file_path, mode) as file:
            file.write(content)
            return file_path

    yield create


class TestRootDirectory:
    def test_zip_file(self, tmpdir, simple_file):
        zip_file = pathlib.Path(f"{tmpdir}/zip_file.zip")
        with zipfile.ZipFile(zip_file, "a") as zip:
            cases = [
                'root_dir/test_file.txt',
                'root_dir/subfolder/test_file.txt',
            ]
            for case in cases:
                file = simple_file(case, '')
                zip.write(file, case)
            zip.writestr(zipfile.ZipInfo(f'empty_dir/'), '')
            zip.close()
        assert root_directory(io.BytesIO(zip_file.read_bytes())) == pathlib.Path('root_dir')

    def test_folders(self, tmpdir, simple_file):
        tmpdir.mkdir("root_dir").join("test_file.txt").write("")
        tmpdir.join("root_dir").mkdir("subfolder").join("test_file.txt").write("")
        assert root_directory(pathlib.Path(tmpdir)) == pathlib.Path(f'{tmpdir}/root_dir')

    def test_exclude(self, tmpdir, simple_file):
        tmpdir.mkdir("root_dir").join("test_file.txt").write("")
        tmpdir.join("root_dir").mkdir("subfolder").join("test_file.txt").write("")
        tmpdir.mkdir("EXCLUDE")
        tmpdir.mkdir("__MACOS").join("test_file.txt").write("")
        assert root_directory(pathlib.Path(tmpdir), exclude=["__MACOS", "EXCLUDE"]) == pathlib.Path(
            f'{tmpdir}/root_dir'
        )



from unittest import TestCase

from fan_tools.testing.meta import PropsMeta


class MetaCase(TestCase, metaclass=PropsMeta):
    def prop__ok(self):
        return True

    def test_simple(self):
        assert self.ok, 'Ok: {!r}'.format(self.ok)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-19 10:30
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import sampleapp.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Article',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('content', models.TextField()),
                ('type', models.IntegerField(null=True, default=sampleapp.models.ArticleType(1))),
            ],
        ),
        migrations.CreateModel(
            name='Author',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('birth_date', models.DateField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Review',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('summary', models.CharField(max_length=255)),
                ('content', models.TextField()),
                ('stars', models.IntegerField()),
                ('article', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reviews', to='sampleapp.Article')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='article',
            name='author',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='articles', to='sampleapp.Author'),
        ),
    ]



#!/usr/bin/env python3
"""
# set DD_KEY, get from https://app.datadoghq.com/organization-settings/api-keys
export DD_KEY=use_your_key_here

# Run vector container
docker run --rm -v $(pwd)/examples/send_dd_metric.toml:/etc/vector/vector.toml -e DD_KEY=$DD_KEY -p 9998:9998/udp timberio/vector:0.14.X-debian
"""
import argparse
import logging

from fan_tools.metrics import send_error_metric, send_metric
from fan_tools.otel.log import enable_otel_logger


logging.basicConfig(level=logging.DEBUG, format='%(asctime)s [%(levelname)s] %(name)s: %(message)s')
log = logging.getLogger('send_dd_metric')


def parse_args():
    parser = argparse.ArgumentParser(description='DESCRIPTION')
    # parser.add_argument('-m', '--mode', default='auto', choices=['auto', 'manual'])
    # parser.add_argument('-l', '--ll', dest='ll', action='store_true', help='help')
    return parser.parse_args()


def main():
    args = parse_args()
    enable_otel_logger(vector_info=('localhost', 9998))

    send_metric('test', tags={'service': '1'})
    # send_error_metric('test_error')
    # send_error_metric('test_error2')


if __name__ == '__main__':
    main()



import pytest
from fan_tools.unix import succ, ExecError


def test_succ_handle_error():
    stderr = []
    with pytest.raises(ExecError) as e:
        succ('bash randomcommand', stderr=stderr)
    assert stderr, 'Should have something: {}'.format(stderr)
    assert e.value.exit_code == 127, e.value



from django.urls import include, re_path
from rest_framework.routers import DefaultRouter

from .views import ArticleViewSet, ReviewViewSet


router = DefaultRouter()
router.register('article', ArticleViewSet)
router.register('review', ReviewViewSet)

urlpatterns = [
    re_path(r'^', include(router.urls)),
]



import asyncio
import time
from unittest.mock import AsyncMock

import pytest

from fan_tools.generic.controller import serialize_merge, Timer
from fan_tools.generic.debug import DebugObject


class TestTimer:
    async def test_basic_functionality(self):
        now = time.time()
        timer = Timer(None)
        await timer.wait(0.03)
        assert pytest.approx(time.time() - now, 0.02) == 0.03

    async def test_trigger(self):
        event_loop = asyncio.get_event_loop()
        now = time.time()
        timer = Timer(None)
        event_loop.call_later(0, timer.trigger)
        await timer.wait(3)
        assert time.time() - now <= 0.1

    async def test_wait_for(self):
        now = time.time()
        timer = Timer(None)
        mm = AsyncMock()
        wf = timer.wait_for(mm(), 3)
        await asyncio.sleep(0.01)
        mm.assert_not_awaited()
        await wf
        assert time.time() - now <= 0.1
        mm.assert_awaited()

    async def test_mutiwait(self):
        timer = Timer(None)
        with pytest.raises(RuntimeError):
            t1, t2 = timer.wait(0.01), timer.wait(0.02)
            await asyncio.gather(t1, t2)
        timer.trigger()

    async def test_join(self):
        now = time.time()
        timer = Timer(None)
        await asyncio.gather(timer.wait(0.01), timer.join())
        assert pytest.approx(time.time() - now, 0.1) == 0.01


class TestDebugObject:
    def test_generic(self):
        g = DebugObject(None)
        c1 = DebugObject(g, name='kube1')
        c1.track('state', 'test')
        DebugObject(g, name='kube2')
        resp = g.serialize()
        for val in resp.values():
            if 'timestamp' in val['values']:
                del val['values']['timestamp']
        assert resp == {
            'global': {'buttons': {}, 'values': {'change_callbacks': 0, 'children': 2}},
            'kube1': {'values': {'state': 'test'}, 'buttons': {}},
            'kube2': {'values': {}, 'buttons': {}},
        }


class TestSerializeMerge:
    def test_empty(self):
        assert serialize_merge({}, {'a': 'b'}, 'some_prefix') == {'a': 'b'}

    def test_prefix(self):
        assert serialize_merge({'values': {'a': 'b'}}, {'values': {}, 'buttons': {}}, 'prefix') == {
            'values': {'prefix.a': 'b'},
            'buttons': {},
        }

    def test_empty_des(self):
        with pytest.raises(KeyError):
            serialize_merge({'values': {'a': 'b'}}, {}, 'prefix')



from fan_tools.django.url import build_absolute_uri


class FakeRequest():

    scheme = 'https'

    def get_host(self):
        return 'google.com'


def test_build_absolute_uri():
    assert build_absolute_uri(FakeRequest()) == 'https://google.com'



import pytest

from fan_tools.django.fields import ChoicesEnum
from fan_tools.python import (
    areduce,
    chunks,
    dict_contains,
    dot_get,
    expand_dot,
    retry,
    slide,
    usd_round,
)


@pytest.fixture(scope='class')
def local_cache():
    pass


@pytest.fixture
def session_settings():
    pass


@pytest.fixture
def foobar_enum():
    class FooBarEnum(ChoicesEnum):
        foo = 1
        bar = 2

    return FooBarEnum


def test_choices_enum(foobar_enum):
    expected = [(1, 'foo'), (2, 'bar')]
    choices = foobar_enum.get_choices()
    assert choices == expected, choices


def test_usd_round():
    assert str(usd_round('12.9800000003')) == '12.98'


def test_04_dict_contains():
    a = {'key': 'aa', 'msg': {'bb': 1, 'cc': True}}
    b = {'key': 'aa', 'msg': {'bb': 2, 'cc': True}}

    assert dict_contains(a, {'key': 'aa'})
    assert dict_contains(b, {'key': 'aa'})
    assert dict_contains(a, {'msg': {'bb': 1}})
    assert dict_contains(b, {'msg': {'bb': 2}})

    assert not dict_contains(a, {'msg': {'bb': 2}})
    assert not dict_contains(b, {'msg': {'bb': 1}})

    assert dict_contains(a, {'msg': {'bb': 1, 'cc': True}})
    assert dict_contains(b, {'msg': {'bb': 2, 'cc': True}})

    assert not dict_contains(a, {'msg': {'bb': 1, 'cc': False}})
    assert not dict_contains(b, {'msg': {'bb': 2, 'cc': False}})
    assert not dict_contains(a, {'msg': {'bb': 1, 'nested': []}})


def test_05_expand_dot():
    assert expand_dot({'msg.payload.status': 'waiting_release'}) == {
        'msg': {'payload': {'status': 'waiting_release'}}
    }


def test_06_slide():
    assert list(slide(range(3))) == [(0, 1), (1, 2), (2, None)]


@pytest.mark.asyncio
async def test_07_async_reduce():
    async def sum_two(a, b):
        return a + b

    assert await areduce(sum_two, [1, 2, 3, 4, 5]) == 15
    assert await areduce(sum_two, [1, 2, 3, 4, 5], initial=100) == 115


def test_08_chunks():
    cases = [
        # test common case
        (list(range(0, 4)), 2, [[0, 1], [2, 3]]),
        # test common case for another n
        (list(range(0, 9)), 3, [[0, 1, 2], [3, 4, 5], [6, 7, 8]]),
        # some cases
        (list(range(0, 10)), 3, [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]),
        (list(range(0, 11)), 3, [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]]),
        # test empty list
        ([], 3, []),
        # test length of list < n
        ([1], 3, [[1]]),
        # test tuple
        ((1, 2, 3, 4, 5, 6), 3, [(1, 2, 3), (4, 5, 6)]),
    ]

    for to_chunks, n, result in cases:
        assert list(chunks(to_chunks, n)) == result


def test_09_retry(capfd):
    @retry(tries=3)
    def func_to_retry():
        print(1)
        raise Exception

    with pytest.raises(Exception):
        func_to_retry()

    out, err = capfd.readouterr()
    assert out == '1\n1\n1\n'


@pytest.mark.asyncio
async def test_09_retry_async(capfd):
    @retry(tries=3)
    async def func_to_retry():
        print(1)
        raise Exception

    with pytest.raises(Exception):
        await func_to_retry()

    out, err = capfd.readouterr()
    assert out == '1\n1\n1\n'


def test_10_retry(capfd):
    @retry(ValueError, tries=3)
    def func_to_retry():
        print(1)
        raise ValueError

    with pytest.raises(ValueError):
        func_to_retry()

    out, err = capfd.readouterr()
    assert out == '1\n1\n1\n'


@pytest.mark.asyncio
async def test_10_retry_async(capfd):
    @retry(ValueError, tries=3)
    async def func_to_retry():
        print(1)
        raise ValueError

    with pytest.raises(ValueError):
        await func_to_retry()

    out, err = capfd.readouterr()
    assert out == '1\n1\n1\n'


def test_11_retry(capfd):
    @retry(tries=3)
    def func_to_retry():
        print(1)

    func_to_retry()

    out, err = capfd.readouterr()
    assert out == '1\n'


@pytest.mark.asyncio
async def test_11_retry_async(capfd):
    @retry(tries=3)
    async def func_to_retry():
        print(1)

    await func_to_retry()

    out, err = capfd.readouterr()
    assert out == '1\n'


def test_12_retry(capfd):
    class A:
        @retry(tries=3)
        def func_to_retry(self):
            print(1)

    A().func_to_retry()

    out, err = capfd.readouterr()
    assert out == '1\n'


@pytest.mark.asyncio
async def test_12_retry_async(capfd):
    class A:
        @retry(tries=3)
        async def func_to_retry(self):
            print(1)

    await A().func_to_retry()

    out, err = capfd.readouterr()
    assert out == '1\n'


def test_13_retry(capfd):
    class A:
        @retry(tries=3)
        def func_to_retry(self, a, b):
            print(a, b)

    A().func_to_retry('111', b='222')

    out, err = capfd.readouterr()
    assert out == '111 222\n'


@pytest.mark.asyncio
async def test_13_retry_async(capfd):
    class A:
        @retry(tries=3)
        async def func_to_retry(self, a, b):
            print(a, b)

    await A().func_to_retry('111', b='222')

    out, err = capfd.readouterr()
    assert out == '111 222\n'


def test_14_dot_get():
    d = {'a': {'b': {'c': [1]}}}
    assert dot_get('a.b.c', d) == [1]
    assert dot_get('a*b*c', d, sep='*') == [1]
    assert dot_get('a*b*c*d', d, sep='*', default=[2]) == [2]
    assert dot_get('a*b*d', d, sep='*', default=[123]) == [123]

    d1 = {'a': None}
    assert dot_get('a.b', d1) is None
    assert dot_get('a.b', d1, None) is None



from django.db import migrations
from fan_tools.django.contrib.postgres.operations import LTreeExtension


class Migration(migrations.Migration):

    dependencies = [
        ('sampleapp', '0001_initial'),
    ]

    operations = [
        LTreeExtension(),
    ]



from enum import IntEnum

from django.conf import settings
from django.db import models

from fan_tools.django.contrib.postgres.models import LTreeModel


class ArticleType(IntEnum):
    article = 1
    review = 2
    ads = 3


class Author(models.Model):
    name = models.CharField(max_length=50)
    birth_date = models.DateField(blank=True, null=True)


class Article(models.Model):
    title = models.CharField(max_length=255)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    content = models.TextField()
    author = models.ForeignKey(Author, on_delete=models.PROTECT, related_name='articles')
    type = models.IntegerField(null=True, default=ArticleType.article)


class Review(models.Model):
    summary = models.CharField(max_length=255)
    content = models.TextField()
    stars = models.IntegerField()
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.DO_NOTHING, null=True, blank=True
    )
    article = models.ForeignKey(Article, on_delete=models.CASCADE, related_name='reviews')


class LTreeModelTest(LTreeModel):
    ltree_defaut_label_field = 'name'
    name = models.CharField(max_length=255)



from pathlib import Path

import pytest

from fan_tools.python import auto_directory, py_rel_path, rel_path
from fan_tools.unix import cd


def test_01_rel_path():
    rel_path('./test_rel_path.py', check=True)
    rel_path('../pyproject.toml', check=True)


def test_02_py_rel_path_deprecated():
    with pytest.deprecated_call():
        py_rel_path('./test_rel_path.py', check=True)
        py_rel_path('../pyproject.toml', check=True)


def test_03_auto_directory(tmpdir):
    with cd(tmpdir):
        auto_directory('./dir_from_string')

    assert (tmpdir / 'dir_from_string').exists

    pth = Path(tmpdir / 'dir_from_path')
    assert not pth.exists(), (pth, tmpdir)
    auto_directory(pth)
    assert pth.exists()



from unittest import mock

import pytest
from asgi_lifespan import LifespanManager

from fan_tools.backup import BackupMonitoring
from fan_tools.backup.utils import get_app, run_main


BS = None


class TestBackup(BackupMonitoring):
    def __init__(self, *args, **kwargs):
        kwargs.pop('bucket_name')
        super().__init__(*args, **kwargs)
        self.monitoring = False
        global BS
        BS = self

    def get_backups_list(self):
        return []

    def perform_upload(self, src_name, dst_name):
        pass

    def run_monitoring(self):
        self.monitoring = True


APP = None


@pytest.fixture
def mock_env(monkeypatch):
    mp = monkeypatch
    mp.setenv('BACKUP_COMMAND', 'sleep 0')

    def _set_app(app, *args, **kwargs):
        print(f'SIDEE: {app=} {args=} {kwargs=}')
        global APP
        APP = app

    with mock.patch('uvicorn.run', side_effect=_set_app):
        yield


@pytest.mark.asyncio
async def test_01_general(mock_env):
    fake_args = mock.MagicMock()
    fake_args.daemonize = True
    fake_args.bucket = 'test_bucket'
    assert not APP
    run_main(TestBackup, args=fake_args)
    assert BS
    assert not BS.monitoring
    async with LifespanManager(APP):
        assert BS.monitoring



# pyright: strict
import asyncio
import functools
import json
import logging
from hashlib import md5
from inspect import isfunction
from pathlib import Path
from types import FunctionType
from typing import Awaitable, Callable, cast, Dict, Generic, Protocol, Type, TypeVar, Union


try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec


default_logger = logging.getLogger('fantools.python')
R = TypeVar('R')
P = ParamSpec('P')


def retry(
    exceptions: Type[Exception] = Exception,
    tries: int = -1,
    logger: logging.Logger = default_logger,
):
    """
    Executes a function and retries it if it failed.

    :param exceptions: an exception or a tuple of exceptions to catch. default: Exception.
    :param tries: the maximum number of attempts. default: -1 (infinite).
    :param logger: will record a call retry warning.
    """

    def decorator(func: Callable[P, R]) -> Union[Callable[P, R], Callable[P, Awaitable[R]]]:
        if asyncio.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                nonlocal tries
                f = functools.partial(func, *args, **kwargs)
                while tries:
                    try:
                        return await f()
                    except exceptions as e:
                        tries -= 1
                        if not tries:
                            raise
                        logger.exception('%s, retrying...', e)
                raise ValueError('tries must be greater than 0')

            return async_wrapper
        else:

            @functools.wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                nonlocal tries
                f = functools.partial(func, *args, **kwargs)
                while tries:
                    try:
                        return f()
                    except exceptions as e:
                        tries -= 1
                        if not tries:
                            raise
                        logger.exception('%s, retrying...', e)
                raise ValueError('tries must be greater than 0')

        return wrapper

    return decorator


class PydanticBaseModel(Protocol):
    def parse_file(self, fname: Union[str, Path]) -> 'PydanticBaseModel':
        ...

    def json(self) -> str:
        ...


ModelType = TypeVar('ModelType', bound=PydanticBaseModel)


class cache_async(Generic[ModelType]):
    """
    file cache for async functions that returns pydantic models
    NB: it doesn't use parameters to generate the cache file name
    """

    def __init__(self, fname: Path, model: ModelType, default: ModelType):
        self.fname = fname
        self._default = default
        self.cache = default
        # load from file
        if self.fname.exists():
            self.cache = model.parse_file(self.fname)

    def __call__(self, func: Callable[P, Awaitable[ModelType]]):
        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> ModelType:
            if self.cache:
                return self.cache
            value = await func(*args, **kwargs)
            self.cache = value
            self.fname.write_text(self.cache.json())
            return value

        wrapper.reset_cache = self.reset_cache  # type: ignore

        return wrapper

    def reset_cache(self):
        self.cache = self._default
        if self.fname.exists():
            self.fname.unlink()


In = ParamSpec('In')
Out = TypeVar('Out')


def memoize(fname: Union[FunctionType, str, Path, None] = None):
    """
    memoize function call to filename
    input and outputs must be JSON serializable
    """

    call_func: Union[FunctionType, None] = None

    if isfunction(fname):
        call_func = fname
        pth = Path('.cache/memoized.json')
    elif not fname:
        pth = Path('.cache/memoized.json')
    else:
        pth = cast(Path, fname)

    pth.parent.mkdir(parents=True, exist_ok=True)

    def reset_cache():
        if pth.exists():
            pth.unlink()

    def _wrapped(func: Callable[In, Out]) -> Callable[In, Out]:
        @functools.wraps(func)
        def _inner(*args: In.args, **kwargs: In.kwargs) -> Out:
            k = md5(json.dumps({'args': args, 'kwargs': kwargs}).encode('utf8'))

            cached: Dict[str, Out] = {}
            if pth.exists():
                cached = json.loads(pth.read_text())

            key = k.hexdigest()

            if key in cached:
                return cached[key]

            res = func(*args, **kwargs)
            cached[k.hexdigest()] = res
            pth.write_text(json.dumps(cached))
            return res

        _inner.reset_cache = reset_cache
        return _inner

    if call_func:
        return _wrapped(call_func)

    return _wrapped



import re
from datetime import date
from unittest import mock

import pytest
from django.db import connection, models, transaction
from rest_framework import status
from rest_framework.reverse import reverse

from fan_tools.django import call_once_on_commit
from fan_tools.django.models import UploadNameGenerator
from pytest_tipsi_django.client_fixtures import UserWrapper

from sampleapp.models import Article, ArticleType, Author, Review


@pytest.fixture
def client(request, module_transaction):
    with module_transaction(request.fixturename):
        return UserWrapper(None)


@pytest.fixture
def author1(request, module_transaction):
    with module_transaction(request.fixturename):
        yield Author.objects.create(name='author 1', birth_date=date(1985, 1, 31))


@pytest.fixture
def author2(request, module_transaction):
    with module_transaction(request.fixturename):
        yield Author.objects.create(name='author 2', birth_date=date(1986, 2, 20))


@pytest.fixture
def articles(request, module_transaction, author1, author2):
    with module_transaction(request.fixturename):
        articles = []
        for i in range(5):
            article_type = i % 3 + 1
            for author in [author1, author2]:
                article = Article.objects.create(
                    title='article {} by {}'.format(i, author.name),
                    content='article content {}'.format(i),
                    author=author,
                    type=article_type,
                )
                articles.append(article)
        Article.objects.create(
            title='null article', content='null content', author=author, type=None
        )
        yield articles


@pytest.fixture
def article_ads(request, module_transaction, author1):
    with module_transaction(request.fixturename):
        yield Article.objects.create(
            title='ads by {}'.format(author1.name),
            content='ads content',
            author=author1,
            type=ArticleType.ads,
        )

@pytest.fixture
def reviews(request, module_transaction, articles, article_ads, author1, author2):
    with module_transaction(request.fixturename):
        reviews = []
        for article in [articles[0], article_ads]:
            for i in range(5):
                review = Review.objects.create(
                    article=article,
                    summary='summary',
                    content='content',
                    stars=5,
                )
                reviews.append(review)
        yield reviews


def test_get_articles_of_review_type(client, articles):
    response = client.get_json(
        reverse('article-list'),
        {'type': 'review'},
        expected=status.HTTP_200_OK,
    )
    data = response['data']
    expected = len([a for a in data if a['type'] == 'review'])
    assert len(data) == expected, data


def test_get_articles_of_ads_and_review_types(client, articles):
    response = client.get_json(
        reverse('article-list'),
        {'type': 'ads,review'},
        expected=status.HTTP_200_OK,
    )
    assert len(response['data']) == 6


def test_null(client, articles):
    response = client.get_json(
        reverse('article-list'),
        {'type': 'null'},
        expected=status.HTTP_200_OK,
        )
    assert len(response['data']) == 1


def test_reviews(client, reviews):
    response = client.get_json(
        reverse('review-list'),
        expected=status.HTTP_200_OK,
    )
    assert len(response['data']) == 10


def test_reviews_empty_filter(client, reviews):
    response = client.get_json(
        reverse('review-list'),
        {'article_type': ''},
        expected=status.HTTP_200_OK,
    )
    assert len(response['data']) == 10


def test_reviews_article_type(client, reviews):
    response = client.get_json(
        reverse('review-list'),
        {'article_type': 'ads'},
        expected=status.HTTP_200_OK,
    )
    assert len(response['data']) == 5


def test_reviews_article_type_err(client, reviews):
    client.get_json(
        reverse('review-list'),
        {'article_type': 'ERROR'},
        expected=400,
    )


def test_reviews_article_type_int(client, articles):
    pk = articles[0].id
    url = reverse('article-detail', [pk])
    assert articles[0].type == ArticleType.article
    data = client.patch_json(
        url,
        {'type': str(ArticleType.ads.value), 'title': 'some title'}
    )['data']
    print(data)
    assert data['type'] == ArticleType.ads.name



@pytest.fixture
def author(db):
    return Author.objects.create(name='author 1', birth_date=date(1985, 1, 31))


@pytest.fixture
def req_cache():
    return {}


@call_once_on_commit
def _test_hook():
    pass


def author_hook(sender, instance, **kwargs):
    _test_hook()


@pytest.fixture
def author_signal(author):
    models.signals.post_save.connect(author_hook, sender=Author)
    yield
    models.signals.post_save.disconnect(author_hook, sender=Author)


def test_once_on_commit(author, req_cache, author_signal):
    with mock.patch(
        'fan_tools.django._get_request_unique_cache',
        side_effect=lambda: req_cache,
    ):
        with transaction.atomic():
            author.birth_date = date(1986, 1, 31)
            author.save()
            author.birth_date = date(1986, 1, 31)
            author.save()

            fn_name = _test_hook.__name__
            call_times = sum(fn_name in str(i[1]) for i in connection.run_on_commit)
            expected_call_times = 1
            assert call_times == expected_call_times

        assert author.birth_date == date(1986, 1, 31)


def test_uploadnamegenerator():
    assert re.match(
        r'static/imageupload/imageupload-image-.*.jpg',
        UploadNameGenerator('imageupload', 'image')(None, 'file.jpg'),
    )



import logging

from contextvars import ContextVar
from datetime import datetime
from typing import Any, Callable, Dict, Optional


log = logging.getLogger(__name__)


class DebugObject:
    def __init__(self, parent, name=None, serialize: Optional[Callable[[Dict], Dict]] = None):
        self.parent = parent
        if self.parent:
            self.parent.add_child(self)
        self.children = []
        self.name = name
        self.serialize_callback = serialize
        self.values = {}
        self.buttons = {}
        self.change_callbacks = []

    def add_child(self, child: 'DebugObject'):
        self.children.append(child)

    def remove_child(self, child):
        if child:
            try:
                self.children.remove(child)
            except ValueError:
                log.warning(f'Child {child} not found in {self.children}')

    def track(self, name, value):
        self.values[name] = value
        self.notify_change()

    def serialize(self) -> Dict[str, Any]:
        out = {}
        if self.name:
            out[self.name] = {
                'values': self.values.copy(),
                'buttons': self.buttons,
            }
            out[self.name]['values']['timestamp'] = datetime.now().isoformat()
            if self.serialize_callback:
                out[self.name] = self.serialize_callback(out[self.name])
            return out
        out['global'] = {
            'values': {
                'change_callbacks': len(self.change_callbacks),
                'children': len(self.children),
            },
            'buttons': {},
        }
        for child in self.children:
            out.update(child.serialize())
        return out

    def add_notify(self, cb):
        self.change_callbacks.append(cb)

    def remove_notify(self, cb):
        try:
            self.change_callbacks.remove(cb)
        except ValueError:
            log.warning(f'Callback {cb} not found in {self.change_callbacks}')

    def __repr__(self) -> str:
        return f'<DebugObject (name={self.name} id={id(self)})'

    def notify_change(self):
        if self.parent:
            return self.parent.notify_change()

        data = self.serialize()
        for cb in self.change_callbacks:
            try:
                cb(data)
            except Exception as e:
                log.exception(f'Exception in debug callback: {e}')


DEBUG_GLOBAL = ContextVar('debug_global', default=DebugObject(None))



from starlette.requests import Request
from starlette.routing import Route, Router
from starlette.testclient import TestClient

from fan_tools.container_utils import HealthView, PrometheusView


async def health_extra(request: Request):
    return {'test': 'OK'}


def prometheus_extra(request: Request):
    return ['test 1']


app = Router(
    [
        Route("/health/", endpoint=HealthView(), methods=["GET"]),
        Route("/health_extra/", endpoint=HealthView(health_extra), methods=["GET"]),
        Route("/prometheus/", endpoint=PrometheusView(), methods=["GET"]),
        Route("/prometheus_extra/", endpoint=PrometheusView(prometheus_extra), methods=["GET"]),
    ]
)

client = TestClient(app)


class TestHealth:
    def test_default(self):
        response = client.get('/health/')
        assert response.status_code == 200
        assert response.json() == {'status': 'OK'}

    def test_extra(self):
        response = client.get('/health_extra/')
        assert response.status_code == 200
        assert response.json() == {'status': 'OK', 'test': 'OK'}


class TestPrometheus:
    def test_default(self):
        response = client.get('/prometheus/')
        assert response.status_code == 200
        assert response.content == b'up 1'

    def test_extra(self):
        response = client.get('/prometheus_extra/')
        assert response.status_code == 200
        assert response.content == b'up 1\ntest 1'



import pytest
from fan_tools.python import rel_path
from fan_tools.image_utils import Transpose


@pytest.fixture
def png_content():
    with rel_path('./image.png').open('rb') as f:
        yield f.read()


def test_png_to_jpg(png_content):
    out = Transpose().process_binary(png_content, raise_on_open=True)
    assert len(out) > 0
    assert out != png_content



import logging
from unittest.mock import MagicMock

from fan_tools.fan_logging import setup_logger


def test_01_js_formatter(tmp_path):
    mm = MagicMock()
    setup_logger('in_test', root_dir=tmp_path, json_params={'get_context': mm})
    mm.assert_not_called()
    logging.debug('test')
    mm.assert_called_once()



import os
import tempfile
import time
from contextlib import suppress

import pytest

from fan_tools.unix import ExecError, succ, wait_socket


def create_database():
    print('Create database')
    c = 0
    while c < 30:
        try:
            succ(
                'echo create database plugin | '
                'PGPASSWORD=password psql -h localhost -p 40001 -U postgres'
            )
            return
        except ExecError as e:
            print(e)
            c += 1
            time.sleep(1)
    print('Cannot create database')
    c, out, err = succ('docker logs testpostgres')
    print('\n'.join(out))
    print('=' * 80)
    print('\n'.join(err))
    exit(1)


def pytest_configure(config):
    if config.getoption('docker_skip'):
        return

    with suppress(Exception):
        print('Stop old container if exists')
        os.system('docker stop testpostgres')
        time.sleep(1)
    with suppress(Exception):
        os.system('docker rm testpostgres')
    try:
        print('Start docker')
        tmp_dir = tempfile.mkdtemp()
        succ(
            'docker run -d -p 40001:5432  -e POSTGRES_PASSWORD=password '
            f'-v {tmp_dir}:/var/lib/postgresql/data '
            '--name=testpostgres postgres:15 '
        )
        time.sleep(5)
        wait_socket('localhost', 40001, timeout=15)
        create_database()
    except Exception as e:
        print('EXCEPTION: {}'.format(e))
        exit(1)


def pytest_unconfigure(config):
    if config.getoption('docker_skip') or config.getoption('keep_db'):
        return
    os.system('docker stop -t 2 testpostgres')
    os.system('docker rm testpostgres')


@pytest.fixture(autouse=True)
def autodatabase(module_transaction):
    pass


def pytest_addoption(parser):
    group = parser.getgroup('fan_group')
    group.addoption(
        '--docker-skip', action='store_true', default=False, help='skip docker initialization'
    )
    group.addoption('--keep-db', action='store_true', default=False, help='skip shutting down db')



from fan_tools.drf.filters import EnumFilter
from fan_tools.drf.serializers import EnumSerializer

import django_filters
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import serializers, viewsets
from sampleapp.models import Article, ArticleType, Review


class ArticleSerializer(serializers.ModelSerializer):
    type = EnumSerializer(ArticleType)

    class Meta:
        model = Article
        fields = ['id', 'title', 'type']


class ArticleFilter(django_filters.FilterSet):
    type = EnumFilter(ArticleType, raise_on_error=True)

    class Meta:
        model = Article
        fields = ['type']


class ArticleViewSet(viewsets.ModelViewSet):
    queryset = Article.objects.all()
    serializer_class = ArticleSerializer
    filter_backends = [DjangoFilterBackend]
    filter_class = ArticleFilter


class ReviewSerializer(serializers.ModelSerializer):
    article = ArticleSerializer()

    class Meta:
        model = Review
        fields = ['id', 'summary', 'content', 'article']


class ReviewFilter(django_filters.FilterSet):
    article_type = EnumFilter(ArticleType, field_name='article__type', raise_on_error=True)

    class Meta:
        model = Review
        fields = ['article_type']


class ReviewViewSet(viewsets.ModelViewSet):
    queryset = Review.objects.all()
    serializer_class = ReviewSerializer
    filter_backends = [DjangoFilterBackend]
    filter_class = ReviewFilter



# Generated by Django 2.2.1 on 2019-06-03 09:18

from django.conf import settings
import django.contrib.postgres.indexes
from django.db import migrations, models
import django.db.models.deletion
import fan_tools.django.contrib.postgres.fields.ltree
import fan_tools.django.contrib.postgres.indexes


class Migration(migrations.Migration):

    dependencies = [
        ('sampleapp', '0002_ltree_extension'),
    ]

    operations = [
        migrations.AlterField(
            model_name='article',
            name='author',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='articles', to='sampleapp.Author'),
        ),
        migrations.AlterField(
            model_name='review',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL),
        ),
        migrations.CreateModel(
            name='LTreeModelTest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ltree_label', fan_tools.django.contrib.postgres.fields.ltree.LTreeLabelField()),
                ('ltree_label_path', fan_tools.django.contrib.postgres.fields.ltree.LTreeLabelPathField()),
                ('name', models.CharField(max_length=255)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='sampleapp.LTreeModelTest')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddIndex(
            model_name='ltreemodeltest',
            index=django.contrib.postgres.indexes.GistIndex(fields=['ltree_label_path'], name='sampleapp_l_ltree_l_5c8efd_gist'),
        ),
        migrations.AddIndex(
            model_name='ltreemodeltest',
            index=django.contrib.postgres.indexes.Index(fields=['ltree_label_path'], name='sampleapp_l_ltree_l_16ae5a'),
        ),
        migrations.AddIndex(
            model_name='ltreemodeltest',
            index=fan_tools.django.contrib.postgres.indexes.LTreeIndex(fields=['ltree_label_path'], name='sampleapp_l_ltree_l_ca7bc1_idx'),
        ),
    ]



from typing import Any, Dict, Type

import pytest

from django.views import View
from fan_tools.drf.validation import validate_request
from rest_framework import serializers, status
from rest_framework.response import Response
from rest_framework.test import APIRequestFactory
from rest_framework.views import APIView
from rest_framework.viewsets import ViewSet

URL = '/myurl'


def get_assigned_request_attrs(view) -> Dict[str, Any]:
    request_attrs = {}
    if hasattr(view, 'validated_query_params'):
        request_attrs['validated_query_params'] = view.validated_query_params
    if hasattr(view, 'validated_data'):
        request_attrs['validated_data'] = view.validated_data
    return request_attrs


@pytest.fixture()
def req_factory():
    yield APIRequestFactory()


@pytest.fixture()
def apiview_cls():
    class MyAPIView(APIView):
        def get(self, request):
            return Response(status=status.HTTP_200_OK, data=get_assigned_request_attrs(self))

        def post(self, request):
            return Response(status=status.HTTP_200_OK, data=get_assigned_request_attrs(self))

    yield MyAPIView


@pytest.fixture()
def viewset_cls():
    class MyViewSet(ViewSet):
        def list(self, request):
            return Response(status=status.HTTP_200_OK, data=get_assigned_request_attrs(self))

        def create(self, request):
            return Response(status=status.HTTP_200_OK, data=get_assigned_request_attrs(self))

    yield MyViewSet


def test_no_error_if_no_data_is_passed_and_no_serializers_defined(
    req_factory: APIRequestFactory, apiview_cls: Type[APIView]
):
    apiview_cls.get = validate_request()(apiview_cls.get)
    request = req_factory.get(URL)
    assert apiview_cls.as_view()(request).status_code == 200


def test_can_only_be_used_on_class_based_apiviews(req_factory: APIRequestFactory):
    request = req_factory.get(URL)
    with pytest.raises(TypeError, match='cannot be used on function-based views'):

        @validate_request()
        def my_view(request):
            pass

        my_view(request)

    request = req_factory.get(URL)
    with pytest.raises(TypeError, match='can only be used on methods of APIView'):

        class MyView(View):
            @validate_request()
            def get(self, request):
                pass

        MyView.as_view()(request)


def test_raises_400_if_query_params_are_invalid(
    req_factory: APIRequestFactory, apiview_cls: Type[APIView]
):
    apiview_cls.get = validate_request(
        query_params={'username': serializers.CharField(allow_blank=False, required=True)}
    )(apiview_cls.get)

    response = apiview_cls.as_view()(req_factory.get(URL))
    assert response.status_code == 400

    response = apiview_cls.as_view()(req_factory.get(URL, data={}))
    assert response.status_code == 400

    response = apiview_cls.as_view()(req_factory.get(URL, data={'username': ''}))
    assert response.status_code == 400


def test_raises_400_if_data_is_invalid(req_factory: APIRequestFactory, apiview_cls: Type[APIView]):
    apiview_cls.post = validate_request(
        data={'username': serializers.CharField(allow_blank=False, required=True)}
    )(apiview_cls.post)

    response = apiview_cls.as_view()(req_factory.post(URL))
    assert response.status_code == 400

    response = apiview_cls.as_view()(req_factory.post(URL, data={}))
    assert response.status_code == 400

    response = apiview_cls.as_view()(req_factory.post(URL, data={'username': ''}))
    assert response.status_code == 400


def test_raises_400_if_data_is_invalid_as_serializer(
    req_factory: APIRequestFactory, apiview_cls: Type[APIView]
):
    class MySerializer(serializers.Serializer):
        username = serializers.CharField(allow_blank=False, required=True)

    apiview_cls.post = validate_request(data=MySerializer)(apiview_cls.post)

    response = apiview_cls.as_view()(req_factory.post(URL))
    assert response.status_code == 400

    response = apiview_cls.as_view()(req_factory.post(URL, data={}))
    assert response.status_code == 400

    response = apiview_cls.as_view()(req_factory.post(URL, data={'username': ''}))
    assert response.status_code == 400


def test_has_query_params_if_passed_and_valid(
    req_factory: APIRequestFactory, apiview_cls: Type[APIView]
):
    apiview_cls.get = validate_request(
        query_params={'username': serializers.CharField(allow_blank=False, required=True)}
    )(apiview_cls.get)

    response = apiview_cls.as_view()(req_factory.get(URL, data={'username': 'maksim'}))
    assert response.status_code == 200
    assert response.data['validated_query_params']['username'] == 'maksim'
    assert 'validated_data' not in response.data


def test_has_data_if_passed_and_valid(req_factory: APIRequestFactory, apiview_cls: Type[APIView]):
    apiview_cls.post = validate_request(
        data={'username': serializers.CharField(allow_blank=False, required=True)}
    )(apiview_cls.post)

    response = apiview_cls.as_view()(req_factory.post(URL, data={'username': 'maksim'}))
    assert response.status_code == 200
    assert response.data['validated_data']['username'] == 'maksim'
    assert 'validated_query_params' not in response.data


def test_400_for_invalid_data_on_viewsets(req_factory: APIRequestFactory, viewset_cls: ViewSet):
    viewset_cls.list = validate_request(
        query_params={'username': serializers.CharField(allow_blank=False, required=True)}
    )(viewset_cls.list)

    request = req_factory.get(URL, data={'username': ''})
    response = viewset_cls.as_view({'get': 'list'})(request)
    assert response.status_code == 400


def test_valid_query_params_on_viewsets(req_factory: APIRequestFactory, viewset_cls: ViewSet):
    viewset_cls.list = validate_request(
        query_params={'username': serializers.CharField(allow_blank=False, required=True)}
    )(viewset_cls.list)

    request = req_factory.get(URL, data={'username': 'maksim'})
    response = viewset_cls.as_view({'get': 'list'})(request)
    assert response.status_code == 200
    assert response.data['validated_query_params']['username'] == 'maksim'
    assert 'validated_data' not in response.data


def test_valid_data_on_viewsets(req_factory: APIRequestFactory, viewset_cls: ViewSet):
    viewset_cls.create = validate_request(
        data={'username': serializers.CharField(allow_blank=False, required=True)}
    )(viewset_cls.create)

    request = req_factory.post(URL, data={'username': 'maksim'})
    response = viewset_cls.as_view({'post': 'create'})(request)
    assert response.status_code == 200
    assert response.data['validated_data']['username'] == 'maksim'
    assert 'validated_query_params' not in response.data



from unittest.mock import patch

import pytest
from asgi_lifespan import LifespanManager
from fastapi import FastAPI
from httpx import AsyncClient

from fan_tools.mon_server import MetricsServer


pytestmark = pytest.mark.asyncio


@pytest.fixture
def server_port(unused_tcp_port_factory):
    yield unused_tcp_port_factory()


@pytest.fixture
async def fastapi_app():
    yield FastAPI()


class MetricHook:
    def __init__(self):
        self.names = {}

    async def remember_update(self, update_func):
        self.update_func = update_func

    def push_metrics(self, dct):
        self.update_func(dct)


@pytest.fixture
def mhook():
    yield MetricHook()


@pytest.fixture
def mserver(fastapi_app, mhook):
    server = MetricsServer(fastapi_app, 'running')
    server.add_task(mhook.remember_update)
    yield server


@pytest.fixture
def patch_loop(event_loop):
    with patch('asyncio.get_event_loop') as m:
        m.return_value = event_loop


@pytest.fixture
async def running_server(fastapi_app, mserver, server_port, patch_loop, event_loop):
    async with LifespanManager(fastapi_app):
        yield fastapi_app


@pytest.fixture
async def get_metrics(running_server, server_port):
    async with AsyncClient(app=running_server) as cli:
        # with TestClient(running_server) as cli:
        async def async_get():
            resp = await cli.get(f'http://localhost:{server_port}/metrics')
            assert resp.status_code == 200
            return resp.text

        yield async_get


@pytest.fixture
def expect_metrics(get_metrics):
    async def inner_expect(expected_metrics={}):
        resp = await get_metrics()
        resp_dict = {}
        for l in resp.split('\n'):
            if not l:
                continue
            k, v = l.split(' ')
            resp_dict[k] = int(v)
        assert resp_dict == expected_metrics

    yield inner_expect


async def test_01(expect_metrics, mhook):
    await expect_metrics({'running': 1})

    mhook.push_metrics({'a': 1})
    await expect_metrics({'a': 1, 'running': 1})

    mhook.push_metrics({'b': 1})
    await expect_metrics({'a': 0, 'b': 1, 'running': 1})



import logging

from fan_tools.metrics import send_error_metric, send_metric


def test_01_metric(caplog):
    tags = {'ttt': 'mmm', 'env': 'prod', 'service': 'cool'}
    with caplog.at_level(logging.INFO, logger='fan.metric'):
        send_metric('meme', tags=tags)
        send_metric('meme2')
    assert len(caplog.records) == 2
    l1 = caplog.records[0]
    assert l1.msg['tags'] == tags
    assert l1.msg['message'] == 'meme'
    l2 = caplog.records[1]
    assert l2.msg['tags'] == {'env': 'not_set', 'service': 'not_set'}


def test_02_error_metric(caplog):
    tags = {'ttt': 'mmm', 'env': 'prod', 'service': 'cool', 'error_type': 'lol'}

    with caplog.at_level(logging.ERROR, logger='fan.metric'):
        send_error_metric('meme', tags=tags)
    assert len(caplog.records) == 1
    l1 = caplog.records[0]
    tags_target = {**tags, 'error_type': 'meme'}
    assert l1.msg['tags'] == tags_target



from sampleapp.models import LTreeModelTest


def test_ltree():
    parent = LTreeModelTest.objects.create(name='parent')
    child = LTreeModelTest.objects.create(name='child', parent=parent)
    grandchild = LTreeModelTest.objects.create(name='grandchild', parent=child)

    assert list(
        LTreeModelTest.objects.filter(
            ltree_label_path__descendants=parent.ltree_label_path
        ).values_list(
            'name', flat=True,
        )
    ) == [
        parent.name,
        child.name,
        grandchild.name,
    ]

    assert list(
        LTreeModelTest.objects.filter(
            ltree_label_path__nlevel=2,
        ).values_list(
            'name', flat=True,
        )
    ) == [
        child.name,
    ]



import asyncio

import pytest
from fan_tools.unix import asucc, ExecError



async def test_asucc(event_loop):
    ret, out, err = await asucc(
        'echo 1; echo 2 >&2; echo 3 >&2; echo 4; sleep 1', check_stderr=False
    )
    assert ret == 0
    assert out == ['1', '4']
    assert err == ['2', '3']

    # Test for out not reused from prev run
    ret, out, err = await asucc(
        'echo 1; echo 2 >&2; echo 3 >&2; echo 4; sleep 1', check_stderr=False
    )
    assert ret == 0
    assert out == ['1', '4']
    assert err == ['2', '3']


async def test_asucc_handle_error(event_loop):
    stderr = []
    with pytest.raises(ExecError) as e:
        await asucc('bash randomcommand', stderr=stderr)
    assert stderr, 'Should have something: {}'.format(stderr)
    assert e.value.exit_code == 127, e.value


async def run_sleep(loop):
    return await asucc('sleep 10000', loop=loop)


@pytest.mark.asyncio
async def test_asucc_kill(event_loop):
    asucc_wait = asyncio.ensure_future(run_sleep(event_loop))
    await asyncio.sleep(0.01)
    asucc_wait.cancel()
    await asyncio.sleep(0.01)
    with pytest.raises(asyncio.CancelledError):
        await asucc_wait



import pytest

from fan_tools.container_utils import MetricStorage


class TestMetricStorage:
    def test_01_init(self):
        storage = MetricStorage()
        assert storage.get_metrics() == {}

    def test_02_increment(self):
        storage = MetricStorage()
        storage.increment('test_key')
        storage.increment('test_key')
        assert storage.get_metrics() == {'test_key': 2}

        storage = MetricStorage()
        storage.increment('test_key_new')
        assert storage.get_metrics() == {'test_key': 2, 'test_key_new': 1}

    def test_03_push(self):
        storage = MetricStorage()
        storage.push('push_metric', 123)
        storage.push('push_metric', 321)
        assert storage.get_metrics() == {
            'test_key': 2,
            'test_key_new': 1,
            'push_metric': [123, 321],
        }

    def test_04_reset(self):
        storage = MetricStorage()
        assert storage.get_metrics() == {
            'test_key': 2,
            'test_key_new': 1,
            'push_metric': [123, 321],
        }

        storage.reset()
        assert storage.get_metrics() == {}

    def test_05_incorrect_types(self):
        storage = MetricStorage()
        storage.push('metric', 123)
        with pytest.raises(ValueError) as exc:
            storage.increment('metric')
        assert str(exc.value) == 'Metric not support incrementing'

        storage.reset()
        assert storage.get_metrics() == {}

        storage.increment('metric')
        with pytest.raises(ValueError) as exc:
            storage.push('metric', 123)
        assert str(exc.value) == 'Metric not support pushing'



import pytest
from fan_tools.testing import ApiUrls


@pytest.fixture(scope='session')
def api_v_base():
    yield 'http://localhost/v001'


@pytest.fixture(scope='session')
def api(api_v_base):
    yield ApiUrls(
        '{}/'.format(api_v_base),
        {
            'password_reset_request': 'password/request/code/',
            'password_reset': 'password/reset/',
            'user_review_list': 'user/{user_id}/review/',
            'user_review': 'user/{user_id}/review/{review_id}/',
            'wine_review': 'wine/{wine_id}/review/',
            'drink_review': 'drink/{drink_id}/review/',
        },
    )


def test_review_list(api):
    user_id = 'user_1'
    assert api.user_review_list(user_id=user_id) == 'http://localhost/v001/user/user_1/review/'
    with pytest.raises(KeyError):
        api.user_review_list()
    assert api.password_reset(user_id=user_id) == 'http://localhost/v001/password/reset/'



import asyncio
import enum
import logging
import time

from asyncio import CancelledError, Future, Task, shield
from typing import Awaitable, Callable, Coroutine, Dict, List, Optional

from fan_tools.generic.debug import DEBUG_GLOBAL, DebugObject


log = logging.getLogger(__name__)


class State(enum.Enum):
    STARTING = 'starting'
    CALLBACK = 'callback'
    SLEEP = 'sleep'
    BACKOFF = 'backoff'
    ON_ERROR = 'on_error'
    STOPPING = 'stopping'
    STOPPED = 'stopped'


class NamedLogger:
    def __init__(self, name):
        self.name = name

    def _log(self, level, msg, *args, **kwargs):
        msg = f'<{self.name}> {msg}'
        getattr(log, level)(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self._log('debug', msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self._log('info', msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        self._log('warn', msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self._log('warning', msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        self._log('exception', msg, *args, **kwargs)


def serialize_merge(src: Dict, dst: Dict, prefix: str) -> Dict:
    for section in ['values', 'buttons']:
        for name, value in src.get(section, {}).items():
            dst[section][f'{prefix}.{name}'] = value
    return dst


class Timer:
    def __init__(self, parent):
        self.parent = parent
        self.wait_event = asyncio.Event()
        self.wait_futures: List[Future] = []
        self._wait_time = 0

    async def wait(self, timeout: int | float):
        if self.has_wait_futures:
            raise RuntimeError('Already waiting')

        wait_futures: List[Future] = [self._sleep(timeout)]
        self._wait_time = time.time() + timeout
        await self._wait_and_clean(wait_futures)

    async def wait_for(self, coro: Coroutine, timeout: int | float):
        if self.has_wait_futures:
            raise RuntimeError('Already waiting')

        f = self._wait_and_clean([self._sleep(timeout), asyncio.create_task(coro)])
        await f

    async def join(self):
        if self.has_wait_futures:
            await asyncio.wait(self.wait_futures, return_when=asyncio.FIRST_COMPLETED)

    def __repr__(self) -> str:
        return f'<Timer wait={self.current_wait_time:.2f}s>'

    @property
    def has_wait_futures(self) -> bool:
        if not self.wait_futures:
            return False
        return any([not t.done() for t in self.wait_futures])

    @property
    def current_wait_time(self) -> float:
        if not self.has_wait_futures:
            return 0
        return max(self._wait_time - time.time(), 0)

    async def _wait_and_clean(self, fs: List[Future]):
        self.wait_futures = fs
        try:
            await self.join()
        finally:
            for f in fs:
                if not f.done() and not f.get_loop().is_closed():
                    f.cancel()
            self.wait_futures = []

    def _sleep(self, timeout: float | int):
        ret = asyncio.create_task(asyncio.sleep(timeout))
        self._wait_time = time.time() + timeout
        return ret

    def serialize(self):
        return {
            'values': {
                'wait_time': self.current_wait_time.__round__(2),
                'wait_futures': len(self.wait_futures),
            }
        }

    def trigger(self):
        if self.has_wait_futures:
            self.wait_futures[0].cancel()


Callback = Callable[..., None] | Callable[..., Awaitable[None]]
ACTIVE_LOOP = 'active_loop'


class Controller:
    stop_states = {State.STOPPING, State.STOPPED}

    def __init__(self, retry_timeout=5, loop_interval=60, loop_timeout=120, debug=False):
        self.state = State.STOPPED
        self.timer = Timer(self)
        self.retry_timeout = retry_timeout
        self.loop_interval = loop_interval
        self.loop_timeout = loop_timeout

        self.active_loops: Dict[str, Task] = {}
        self.on_stop_callbacks: Dict[str, Callback] = {}

        self.debug: Optional[DebugObject] = None
        self.name = getattr(self, 'name', f'{self.__class__.__name__}_{id(self)}')
        self.log = NamedLogger(self.name)

        if debug:
            if isinstance(debug, DebugObject):
                self.debug = DebugObject(parent=debug, name=self.name, serialize=self.serialize)
            else:
                self.debug = DebugObject(
                    parent=DEBUG_GLOBAL.get(), name=self.name, serialize=self.serialize
                )

            if self.debug:

                def _remove_child():
                    self.debug.parent.remove_child(self.debug)

                self.add_stop_callback(f'debug_{self.name}', _remove_child)

    async def loop_iteration(self):
        """
        main loop work is here
        """

    async def startup(self):
        """
        run before the main loop
        """

    async def shutdown(self):
        """
        run after the main loop
        """

    async def on_error(self, exception) -> Optional[bool]:
        """
        return True if shoult stop
        do reinitialization here if required
        """
        self.log.exception(f'Exception in loop: {self} {exception}')

    def serialize(self, out):
        """
        {'values': {}, 'buttons': {}}
        """
        timer_out = self.timer.serialize()
        out = serialize_merge(timer_out, out, 'timer')
        return out

    def add_stop_callback(self, name: str, callback: Callback):
        """
        add dynamic function to run during stop
        """
        if name in self.on_stop_callbacks:
            raise RuntimeError(f'Callback {name} already registered')
        self.on_stop_callbacks[name] = callback

    def track_task(self, task: Task | Coroutine, name: str | None = None):
        """
        register additional loops, they will stop automatically
        """
        if not isinstance(task, Task):
            task = asyncio.create_task(task)

        if not name:
            name = f'task_{id(task)}_{str(task)}'

        if name in self.active_loops:
            raise RuntimeError(f'Task {name} already registered')

        self.active_loops[name] = task

        def __maybe_remove(*_):
            if name in self.active_loops and self.active_loops[name] == task:
                del self.active_loops[name]

        task.add_done_callback(__maybe_remove)

    def set_state(self, value):
        self.log.debug(f'{self} state changed: {self.state} -> {value}')
        self.state = value
        if self.debug:
            self.debug.track('state', str(value))

    def get_retry_timeout(self):
        return self.retry_timeout

    def get_loop_interval(self):
        return self.loop_interval

    async def start(self):
        if self.state != State.STOPPED:
            raise RuntimeError(f'Cannot start {self} in state {self.state}')

        self.state = State.STARTING
        await self.startup()
        active_loop = asyncio.ensure_future(self.inner_loop())
        self.track_task(active_loop, ACTIVE_LOOP)
        return active_loop

    def trigger_stop(self):
        self.log.debug(f'Triggering stop {self}')

        for name, loop in self.active_loops.items():
            self.log.debug(f'Loop item {name} {loop=}')
        for name, callback in self.on_stop_callbacks.items():
            self.log.debug(f'Callback item {name} {callback=}')

        if self.is_running:
            self.set_state(State.STOPPING)
            self.timer.trigger()

    async def stop(self):
        self.trigger_stop()
        active_loop = self.active_loops.get(ACTIVE_LOOP)
        if active_loop:
            await active_loop

    @property
    def is_running(self):
        return self.state not in self.stop_states

    async def inner_loop(self):
        try:
            while self.is_running:
                try:
                    self.set_state(State.CALLBACK)
                    await self.timer.wait_for(self.loop_iteration(), timeout=self.loop_timeout)
                    self.set_state(State.SLEEP)
                    await self.timer.wait(self.get_loop_interval())
                except CancelledError:
                    self.log.debug(f'Cancelled loop: {self}')
                    break
                except Exception as e:
                    self.log.debug('check should_stop')
                    should_stop = await self._call_on_error(e)
                    if should_stop:
                        log.info(f'Stopping loop because of error: {self}')
                        break
                    self.set_state(State.BACKOFF)
                    timeout = self.get_retry_timeout()
                    self.log.info(f'Retry in {timeout} seconds')
                    await self.timer.wait(timeout)
        finally:
            self.log.debug(f'Shielded shutdown in finally {self.state=}')
            await self._do_shutdown()

    async def _call_on_error(self, error) -> bool:
        self.set_state(State.ON_ERROR)
        should_stop = False
        try:
            should_stop = bool(await self.on_error(error))
        except Exception as e:
            self.log.exception(f'Exception in on_error handler: {e}')
        return should_stop

    async def _do_shutdown(self):
        if active_loop := self.active_loops.get(ACTIVE_LOOP):
            self.log.debug(f'is_cancelled={active_loop.cancelled()}')
        self.set_state(State.STOPPING)

        await self.run_shielded_shutdown()
        await self.stop_tracked_tasks()
        await self.run_on_stop_callbacks()
        self.set_state(State.STOPPED)

    async def run_shielded_shutdown(self):
        try:
            self.log.info(f'shielded shutdown {self=}')
            await shield(self.shutdown())
        except Exception:
            self.log.exception('During shielded shutdown')

    async def stop_tracked_tasks(self):
        wait_tasks = []
        for task_name, task in self.active_loops.items():
            if task_name != ACTIVE_LOOP and not task.done():
                self.log.debug(f'is_cancelled={task.cancelled()}')
                task.cancel()
                wait_tasks.append(task)
        try:
            await asyncio.gather(*wait_tasks, return_exceptions=True)
        except Exception:
            self.log.exception('During stopping active tasks')

    async def run_on_stop_callbacks(self):
        for callback_name, callback in self.on_stop_callbacks.items():
            try:
                if asyncio.iscoroutinefunction(callback):
                    await shield(callback())
                else:
                    callback()
            except Exception as e:
                log.exception(f'Exception in on_stop callback={callback_name}: {e}')

    def __repr__(self):
        return f'<{self.__class__.__name__} {id(self)} state={self.state}>'
